:- module(identifier_built_of_terms, [
          splitAtomAtDots/2,              % (Identifier, ListOfLowerCaseTerms)
          identifier_built_of_terms/2     % (Identifier, ListOfLowerCaseTerms)
          ] ).



/**
 * splitAtomAtDots(+AtomWithDots, -ListOfLowerCaseAtoms )
 *
 * Arg1 is an atom containing dots (".").
 * Arg2 is a list of the subatoms of Arg1 separated by dots. 
 * All elements of Arg2 are in lower case. 
 * Example: Arg1 = 'BLA.bLub.FOo'
 *   yields Arg2 = [bla, blub, foo].
 */
splitAtomAtDots(Atom,ListOfLowerCaseAtoms) :-
     downcase_atom(Atom,LowerAtom),
     concat_atom(ListOfLowerCaseAtoms,.,LowerAtom).     

    
/**
 * identifier_built_of_terms(+Identifier, ?Termlist)
 *
 * Arg1 is an atom in CamelCase spelling or containing  
 *      punctuation symbols (such as "." or "_").
 * Arg2 is a list of atoms in lower-case spelling. Each 
 *      atom in the list corresponds to a portion of Arg1
 *      up to (but excluding) the next capital letter or 
 *      punctuation symbol. The order of atoms in the list
 *      is the same as the order in Arg1. 
 * Example: Arg1 = 'thisIsAMix_of_CAMELCase.withPunctuation' yields
 * Arg2 = [this, is, a, mix, of, camel, case, with, punctuation]  
 */
identifier_built_of_terms(Identifier, ListOfAtoms):-
    atom_codes(Identifier, ListOfCharacters),
    splitListAtCapitalLettersAndPunctuationSymbols(ListOfCharacters, ListOfAtoms).


/**
 * splitListAtCapitalLettersAndPunctuationSymbols(+CharacterList, ?AtomList )
 *
 * Arg2 is the a list of lists, each containing just lower-case 
 * letters. The lists are split at the positions where Arg1 
 * contained an uppercase letter. A sequence of upper case letters
 * in Arg1 yields a single sublist of Arg2.
 */
splitListAtCapitalLettersAndPunctuationSymbols([Head|Tail], DowncaseAtomList ):-
     splitList___([Head|Tail], [], ListOfCharlists),
     convert2lowercase(ListOfCharlists, DowncaseAtomList).  % Downcase all atoms in the list


convert2lowercase([],[]).
convert2lowercase([[]|CharLists],LowerCaseAtom) :- 
    convert2lowercase(CharLists,LowerCaseAtom).
convert2lowercase([[A|B]|CharLists],[LowerCaseAtom|LowerCaseAtoms]) :-    
    atom_chars(Atom,[A|B]),               
    downcase_atom(Atom,LowerCaseAtom),
    convert2lowercase(CharLists,LowerCaseAtoms).           
     

/**
 * splitList___(+CharacterList, +Atom, ?AtomLists )
 * 
 * Arg1 is a list of characters including upper case letters
 *      and punctuation symbols (that is, characters for which 
 *      char_type(Char, punct) succeeds). 
 * Arg2 is a list of characters that are not punctuation symbols.
 *      This is just an accumulator. In any non-recursive call 
 *      it should be initialized with the empty list. 
 * Arg3 is the a list of atoms in lower-case letters.
 *      Each atom in the list corresponds to a portion of Arg1
 *      up to (but excluding) the next capital letter or 
 *      punctuation symbol. Punctuation symbols do not appear
 *      in Arg3. A maximal sequence of N consequtive upper case 
 *      letters from Arg1 yields an atom of length N-1. The order 
 *      of atoms in the list is the same as the order in Arg1.
 * Example: Arg1 =  [A,M,i,x,_,o,f,_,_,.,_,C,A,M,E,L,C,a,s,e]' 
 * yields Arg3 = [a, mix, of, camel, case]  
 */ 
splitList___([], Upper, [[], RevUpper] ) :-
    reverse(Upper,RevUpper). 


   
splitList___([Head|Tail], Upper, ListOFCharLists) :-    
    char_type(Head, lower)                                   % if Head has lower case
    -> ( keepLastUpper(Upper, Head, Rest, ListOFCharLists),  % add it to current list after prev. upper
         splitList___(Tail, [], Rest) 
        ).

splitList___([Head|Tail], Upper, ListOFCharLists) :-    
    char_type(Head, digit),                                   % if Head has lower case
    keepAllUpper(Upper, Rest, ListOFCharLists),
    splitList___(Tail, [Head], Rest).
    
 
               
splitList___([Head|Tail], Upper, Res) :-  
    char_type(Head, upper)                              % if Head has upper case
    -> splitList___(Tail, [Head|Upper], Res).           % add it to the Upper list
    
splitList___([Head|Tail], Upper, ListOFCharLists) :-
    char_type(Head, punct),                             % if Head is a punctuation symbol 
    keepAllUpper(Upper, Rest, ListOFCharLists),         % skip it 
    splitList___(Tail, [], Rest).                  



/**
 * keepLastUpper(+Uppers, +Lower, ?ListOFCharListsIn, ?ListOFCharListsOut )
 *
 * Arg1 is a list of upper case characters in the reverse order of their 
 *      occurrence in the initially analysed atom.
 * Arg2 is the first lower case character read after the sequence of uppers.
 * Arg3 is the list of character lists resulting from the analysis of the
 *      rest of the atom.
 * Arg4 ist the total result. If Arg1 is empty, Arg4 is the extension of the
 *      first sublist of Arg3 by inserting Arg2 as its first element.
 *      Otherwise, Arg4 is the extension of the first sublist of Arg3 
 *      by inserting the first element of Arg1 and Arg2 (in this order)
 *      as its first elements. In addition, the reversed rest of Arg1 
 *      becomes the new first element of the result list.
 */
keepLastUpper([],                Head, [First|More], [               [       Head|First] | More ]) .
keepLastUpper([Upper          ], Head, [First|More], [ [],           [Upper, Head|First] | More ]) .
keepLastUpper([Upper|MoreUpper], Head, [First|More], [ [], RevUpper, [Upper, Head|First] | More ]) :-
   MoreUpper \== [],
   reverse(MoreUpper,RevUpper).

/**
 * Arg3 is the list whose first element is the reverse of Arg1 and whose 
 * tail is Arg2
 */
keepAllUpper(UpperList, Rest, [RevUpper|Rest]) :-
   reverse(UpperList,RevUpper).   


    
/* Previous version, without joining of capital letters into a single term:

splitListAtCapitalLettersAndPunctuationSymbols(CharacterList, DowncaseAtomList )
     CharacterList = [Head|Tail],
     %char_type(HeadAtomLower, to_lower(Head)),
     splitList___([Head|Tail], [], ListOfCharlists),
%     ListOfCharlists=DowncaseAtomList.
     maplist(atom_chars, ListOfAtoms, ListOfCharlists),      % Turn character lists into atoms
     maplist(downcase_atom, ListOfAtoms, DowncaseAtomList).  % Downcase all atoms in the list
     
     
splitList___([],  Prev, [Prev] ) .
 
splitList___([Head|Tail], Prev, Rest) :-
    char_type(Head, lower),
    char_type(HeadAtom, to_lower(Head)),
    atom_concat(Prev,HeadAtom,InitialPart),
    splitList___(Tail, InitialPart, Rest).
    
splitList___([Head|Tail], InitialPart, [InitialPart|Rest]) :-
    char_type(Head, upper),
    char_type(HeadAtomLower, to_lower(Head)),
    splitList___([HeadAtomLower|Tail], '', Rest).
    
splitList___([Head|Tail], InitialPart, [InitialPart|Rest]) :-
    char_type(Head, punct),
    splitList___(Tail, '', Rest).
*/    